# -*- coding: utf-8 -*-
"""
@author: Chris Willert, 
    DLR Institute of Propulsion Technology, 51170 Koeln, Germany
    email: chris.willert(at)dlr.de

Purpose:
    performs image back-projection to undistort raw PIV images onto a common grid 
    Uses nonlinear projective mapping as described in Sec.8.3.1
    (see also mapping_demo.py)
    
    - load in image-object point correspondances (previously determined)
    - compute mappings: image-to-object and object-to-image
    - back-project (undistort) images and show result images
    
    The total data set consists of 11 PIV image sets (4 images each)
    including two sets of quiescent flow conditions
    
    for details on the experiment consult:
        PIV-Book, 3rd Edition, Sec.17.1 (p.585ff)
        https://doi.org/10.1088/0957-0233/8/12/010
    
"""

import os
import numpy as np
import matplotlib.pyplot as plt

from mapping_demo import readCalibPoints, calculateMapping,backprojectImage,showImage

import imageio as iio
def loadTIFFImage(fnIN):
    # load image using imageio
    return np.array(iio.imread(fnIN), dtype=float) 
def saveTIFFImage8bit(fnIN, imgIN):
    # save image using imageio
    # convert to 8-bit, including range clip
    img = imgIN.copy()
    img[img > 255] = 255
    img[img < 0] = 0
    return np.array(iio.imwrite(fnIN, img.astype(np.uint8)), dtype=float) 

#%% ------ Main() -----------
if __name__ == '__main__':

    dirIN = '../data/'
    imgDirIN = '../images/'
    imgDirOUT = '../mapped/'    # place to put the mapped images
    plotDirOUT = '../plots/'    # place to some plot output
    for dirOUT in [imgDirOUT,plotDirOUT]:
        if not os.path.exists(dirOUT):
            os.makedirs(dirOUT)
        
    mappingType = 'NonLin1st'
    # mapping = 'NonLin2nd'
    
    mapRev = []
    views = [1,2]
    
    # load correspondances points and compute mapping
    for view in views:
        fn = dirIN + 'calib_pts_view' + str(view) + '.dat'
        if not os.path.exists(fn):
            raise FileNotFoundError(fn)
        
        pts = readCalibPoints(fn, verbose=1)
        mapRev.append(
            calculateMapping(pts, mapping=mappingType, reverse=True, verbose=1))  

    #%% use maps for image back-projection 
    # define size and global magnification factor of output image
    mapParams = {
        'sf' : 8.5,    # [pixel/mm]
        'offset' : [63,52],    # [mm]
        'szImage' : [1024,1280],    # [h,w]
        }

    imgStubs = ['0001','0002', # pulse delay 300us
                '1001','1002', # pulse delay 500us
                '2001',         # pulse delay 400us
                '3001','4001','5001','6001','7001','8001', # pulse delay 200us
                ]
    for i,view in enumerate(views):
        baseName = imgDirIN + 'cam' + str(view) + '/cam' + str(view) + '_'
        destBaseName = imgDirOUT  + 'mapped_cam' + str(view) + '_'
        for stub in imgStubs:
            for suffix in ['a.tif', 'b.tif']:
                fnImg = baseName + stub + suffix
                fnImgOUT = destBaseName + stub + suffix
                if os.path.isfile(fnImg):
                    img = loadTIFFImage(fnImg)
                    imgMapped = backprojectImage(img, mapRev[i], mapParams, ) #method='slinear')
                    showImage(imgMapped)
                    plt.show()
                    plt.close()
                    print('Back-projected image stored in: ' + fnImgOUT)
                    saveTIFFImage8bit(fnImgOUT, imgMapped)

