# -*- coding: utf-8 -*-
"""

@author: Chris Willert, 
    DLR Institute of Propulsion Technology, 51170 Koeln, Germany
    email: chris.willert(at)dlr.de

Purpose:
    computes mapping from image to object space and back using nonlinear projective 
    mapping described in Sec.8.3.1
    
    - load in image-object point correspondances (previously determined)
    - load calibration images
    - plot image with detected calibration markers
    - compute mappings: image-to-object and object-to-image
    - back-project (undistort) image and plot result image
    
    Mappings can be used for image back-projection in stereo PIV processing
    (see map_images.py)


    for details on the experiment consult:
        PIV-Book, 3rd Edition, Sec.17.1 (p.585ff)
        https://doi.org/10.1088/0957-0233/8/12/010
"""

import numpy as np
import os
from scipy.optimize import curve_fit
import matplotlib.pyplot as plt

# newer releases use the following
#import imageio.v3 as iio
import imageio as iio
def loadTIFFImage(fnIN):
    # load image using imageio
    return np.array(iio.imread(fnIN), dtype=float) 


def readCalibPoints(
        fnIN,
        verbose=0,
        ):
    # read point correspondences from ASCII file
    # list of [imgX, imgY, objX, objY, objZ]
    # with image coordinates in pixel, object coordinates in [mm]
    if verbose:
        print('Loading: ' + str(fnIN))
    pts  = [] # start with empty array
    with open(fnIN, 'r') as f:
        for line in f:
            #print(repr(line))
            tokens = line.strip().split() # also remove trailing line feed
            if('#' in line):
                # most likely a comment
                pass
            elif len(tokens) == 5:
                # high likelihood this is a calibration point
                pts.append(np.array(tokens).astype(float))
    
    if verbose:
        print(' --> read %d calibration points' %(len(pts)))
    # finally convert to array of shape [npts,5]
    return np.stack(pts)    

def perspectiveProjection(X_in, *matA) :
    # perspective projection of first order, Eqns 8.15/8.16
    # input: coordinates X_in, coefficient matrix A
    # output: mapped coordinate, either X or Y
    # we use a helper index to distinguish betwen Xand Y coordinate in the mapped coordinate
    # implements Eqns 8.15, 8.16 
    [x,y,idx] = X_in
    [a11,a12,a13, 
     a21,a22,a23, 
     a31,a32,a33 ] = matA
    denom = a31*x + a32*y + 1

    retValX = (a11*x + a12*y + a13)/denom
    retValY = (a21*x + a22*y + a23)/denom

    # single coordinate
    if(X_in.size == 3):
        return retValY if idx>0 else retValX

    isY = np.where(idx>0)
    retValX[isY] = retValY[isY]
    
    return retValX

def perspectiveProjection2ndOrder(X_in, *matA) :
    # perspective projection of first order, Eqns 8.15/8.16
    # input: coordinates X_in, coefficient matrix A
    # output: mapped coordinate, either X or Y
    # we use a helper index to distinguish betwen Xand Y coordinate in the mapped coordinate
    # implements Eqns 8.17, 8.18
    [x,y,idx] = X_in
    [a11,a12,a13,a14,a15,a16, 
     a21,a22,a23,a24,a25,a26, 
     a31,a32,a33,a34,a35,a36 ] = matA
    denom = a31*x + a32*y + 1 + a34*x*x + a35*y*y + a36*x*y

    retVal = (a11*x + a12*y + a13 + a14*x*x + a15*y*y + a16*x*y)/denom
    retValY = (a21*x + a22*y + a23 + a24*x*x + a25*y*y + a26*x*y)/denom
    
    # single coordinate
    if(X_in.size == 3):
        return retValY if idx>0 else retVal

    isY = np.where(idx>0)
    retVal[isY] = retValY[isY]
    
    return retVal

def mapPointsPerCoordinate(ptsIN, mapping):
    # input of type [xi,yi,index_i]
    if mapping['mapping'].lower() == 'nonlin1st':
        return perspectiveProjection( ptsIN, *mapping['coeffs'])
    elif mapping['mapping'].lower() == 'nonlin2nd':
        return perspectiveProjection2ndOrder( ptsIN, *mapping['coeffs'])
    raise ValueError('Mapping type not available: ' + mapping['mapping'])
    
def mapPoints(ptsIN, mapping):
    if ptsIN.size == 2:
        # single point
        npt = 1
        pts = np.array([
            ptsIN[0],ptsIN[1],0,
            ptsIN[0],ptsIN[1],1
            ]).reshape(2,3)
    else:
        # rearrange to suit mapping function
        npt,nc = ptsIN.shape 
        pts = np.zeros([npt*2,3])
        pts[npt:,2] = 1
        pts[:npt,0] = ptsIN[:,0] 
        pts[:npt,1] = ptsIN[:,1] 
        pts[npt:,0] = ptsIN[:,0] 
        pts[npt:,1] = ptsIN[:,1] 
    
    ptsOUT = mapPointsPerCoordinate(pts.T, mapping)
    if npt == 1:
        return ptsOUT.flatten()
    return ptsOUT.reshape(npt,2).T

def calculateMapping(
        calibPtsIN,
        mapping='Nonlin1st',
        reverse=False,
        verbose=0,
        ):
    # compute a mapping between physical and image space
    # mapping types:
    #   NonLin1st - nonlinear projective mapping of first order
    #   NonLin2nd - nonlinear projective mapping of second order
    # reverse: if True performs mapping from image to object space (at Z=0)
    
    # in order to do the fit, we use a helper index to distinguish between X and Y coordinate
    # therefore the data must be shuffled around a bit
    Xo = calibPtsIN[:,0:4].copy() # deep copy required, otherwise overwrites data
    Yo = calibPtsIN[:,0:4].copy()
    Xo[:,2] = 0   # helper index
    Yo[:,2] = 1   # helper index
    if reverse:
        Xo[:,0:2] = calibPtsIN[:,2:4]
        Xo[:,  3] = calibPtsIN[:,0]
        Yo[:,0:2] = calibPtsIN[:,2:4]
        Yo[:,  3] = calibPtsIN[:,1]
    else:    
        Xo[:,3] = calibPtsIN[:,2]
    
    imgData = np.vstack([Xo[:,0:3], Yo[:,0:3]])
    objData = np.vstack([Xo[:,3], Yo[:,3]]).flatten()    
    unit = 'pixel' if reverse else 'mm'
    
    if not mapping.lower() in ['nonlin1st', 'nonlin2nd'] :
        raise ValueError('Unsupported mapping type: ' + mapping)
    
    if mapping.lower() in ['nonlin1st', 'nonlin2nd']:
        # Initial guesses to the fit parameters.
        matA_guess = [
            [ 1, 0, 0],
            [ 0, 1, 0],
            [ 0 ,0, 1]              
            ]
        # Flatten the initial guess parameter list
        matA0 = [a for Aj in matA_guess for a in Aj]   
        
        popt, pcov = curve_fit(
            perspectiveProjection, 
            imgData.T, 
            objData, 
            p0=matA0,
            )        
        # calculate residual to the fit
        mappedObjPts = perspectiveProjection( imgData.T, *popt)
        
    if mapping.lower() in ['nonlin2nd']:
        # matA_guess = [
        #     [ 1, 0, 0, 0, 0, 0],
        #     [ 0, 1, 0, 0, 0, 0],
        #     [ 0 ,0, 1, 0, 0, 0]              
        #     ]
        # use result from 1st fit for initialization of the fit parameters.
        matA_guess = [
            [ popt[0], popt[1], popt[2], 0, 0, 0],
            [ popt[3], popt[4], popt[5], 0, 0, 0],
            [ popt[6] ,popt[7],       1, 0, 0, 0]              
            ]
        # Flatten the initial guess parameter list
        matA0 = [a for Aj in matA_guess for a in Aj]   
        
        popt, pcov = curve_fit(
            perspectiveProjection2ndOrder, 
            imgData.T, 
            objData, 
            p0=matA0,
            )        
        # calculate residual to the fit
        mappedObjPts = perspectiveProjection2ndOrder( imgData.T, *popt)
        
    # residuals of fit
    resid = np.sqrt(np.mean((mappedObjPts - objData)**2)) 
    maxAbsoluteDev = np.sqrt(((mappedObjPts - objData)**2).max())
    if verbose:
        print('Mapping type: ' + mapping 
              + (' (object to image)' if reverse else ' (image to object)'))
        print(' --> mean residual: %g %s  (max. deviation: %g %s)' % (resid, unit, maxAbsoluteDev, unit))
    
    return {
        'mapping' : mapping,
        'coeffs'  : popt,
        'reverse' : reverse,
        }


from scipy.interpolate import RegularGridInterpolator

def backprojectImage(
        imgIN, 
        mapping, 
        mapParams,
        method='linear'  # interpolation method, or ['slinear','cubic','quintic']
        ):

    # create interpolator from input image
    iy = np.arange(imgIN.shape[0])
    ix = np.arange(imgIN.shape[1])
    interp = RegularGridInterpolator(
        (ix, iy), 
        np.transpose(imgIN),
        method=method,
        fill_value=0,
        )
    
    # now scan through destination pixel coordinates, back-project into original
    # image and interpolate the intensity
    imgOUT = np.zeros(mapParams['szImage'])
    ny,nx = imgOUT.shape
    ox = (np.arange(nx) / mapParams['sf']) - mapParams['offset'][0]
    lineOfCoords = np.zeros([imgOUT.shape[1]*2,3]) 
    lineOfCoords[:nx,0] = ox
    lineOfCoords[nx:,0] = ox
    lineOfCoords[nx:,2] = 1
    
    imgIN_ny,imgIN_nx = imgIN.shape
    
    for iy in np.arange(ny):
        # caution: coordinate switch on Y going from image to object space
        oy = ((np.zeros([nx]) + ny - iy)/ mapParams['sf']) - mapParams['offset'][1]
        lineOfCoords[:nx,1] = oy
        lineOfCoords[nx:,1] = oy
        
        # apply mapping: object to original image
        mappedPts = perspectiveProjection( lineOfCoords.T, *mapping['coeffs'])
        mappedXY = mappedPts.reshape(2,nx)
        pts_inside = np.where(
            (mappedXY[0] >= 0) 
            & (mappedXY[0] < imgIN_nx-1)
            & (mappedXY[1] >= 0) 
            & (mappedXY[1] < imgIN_ny-1))
        #print('j=%g inside: %d ' %(iy, pts_inside[0].size) )
        
        if pts_inside[0].size > 0:
            pixOut = interp(mappedXY.T[pts_inside])
            imgOUT[iy,pts_inside] = pixOut  # put in interpolated pixels
           
    return imgOUT

def showImage(
        img,
        figsize=(7,6),
        cmap='gray',
        saveFile = None,
        ):
    fig, ax1 = plt.subplots(1,1, figsize=figsize, tight_layout=True )
    ax1.imshow(img, cmap=cmap)
    ax1.set_xlabel('x  [pixel]')
    ax1.set_ylabel('y  [pixel]')
    if (saveFile is not None):
        plt.savefig(saveFile, dpi=200)
    return ax1

def showImageAndCalibrationPoints(
        img, pts,
        figsize=(7,6),
        cmap='gray',
        ):
    fig, ax1 = plt.subplots(1,1, figsize=figsize, tight_layout=True )
    ax1.imshow(img, cmap=cmap)
    x = pts[:,0].flatten()
    y = pts[:,1].flatten()
    ax1.plot(x,y, 'r+')
    ax1.set_xlabel('x  [pixel]')
    ax1.set_ylabel('y  [pixel]')
    return ax1

#%% ------ Main() -----------
if __name__ == '__main__':

    dirIN = '../data/'
    imgDirIN = '../images/'
    plotDirOUT = '../plots/'    # place to some plot output
    if not os.path.exists(plotDirOUT):
        os.makedirs(plotDirOUT)
    
    mapping = 'NonLin1st'
    # mapping = 'NonLin2nd'
    fnCalibPts = []
    fnCalibImgs = []
    views = [1,2]
    # reqd files - check their existence
    for view in views:
        fn = dirIN + 'calib_pts_view' + str(view) + '.dat'
        if not os.path.exists(fn):
            raise FileNotFoundError(fn)
        fnCalibPts.append(fn)

        fn = imgDirIN + 'cam' + str(view) + '/cam' + str(view) + '_grid_0001.tif'        
        if not os.path.exists(fn):
            raise FileNotFoundError(fn)
        fnCalibImgs.append(fn)
    
    calibImgs = []
    mapFwd = []
    mapRev = []
    for i,fn in enumerate(fnCalibPts):
        # load calibration points    
        pts = readCalibPoints(fn, verbose=1)
    
        img = loadTIFFImage(fnCalibImgs[i])
        calibImgs.append(img)
        
        # plot image with calibration points
        ax = showImageAndCalibrationPoints(img, pts)
        
        mapFwd.append(
            calculateMapping(pts, mapping=mapping, verbose=1))
        mapRev.append(
            calculateMapping(pts, mapping=mapping, reverse=True, verbose=1))  
        
        # mark the origin
        ptOrigin = mapPoints(np.array([0,0]), mapRev[i])        
        ax.plot(ptOrigin[0], ptOrigin[1], '+', ms=10, color='#00ff00')
        
        # mark major axes
        somePoints = mapPoints(np.array([[30,0],[0,30]]), mapRev[i])
        for pt in somePoints:
            ax.plot(pt[0], pt[1], 'o', ms=8, color='#ffff00', mfc='None')
            
        plt.savefig(plotDirOUT + 'calib' + str(i+1) + '_with_markers.png', dpi=200)
        
        plt.show()
        plt.close()

    #%% use maps for image back-projection 
    # define size and global magnification factor of output image
    mapParams = {
        'sf' : 8.5,    # [pixel/mm]
        'offset' : [63,52],    # [mm]
        'szImage' : [1024,1280],    # [h,w]
        }
    
    mappedImgs = []
    for i,view in enumerate(views):
        
        imgMapped = backprojectImage(calibImgs[i], mapRev[i], mapParams, ) #method='slinear')
        mappedImgs.append(imgMapped)
        
        showImage(imgMapped, cmap='gray',
                  saveFile=plotDirOUT + 'mapped_calib' + str(view) + '.png')

        plt.show()
        plt.close()
        
    # overlay both mapped images
    showImage(mappedImgs[0] + mappedImgs[1], cmap='gray',
              saveFile=plotDirOUT + 'mapped_overlaid.png')
    plt.show()
    plt.close()


